@@ -1,6 +1,5 @@
-# -*- coding: utf-8 -*-
 #
-# Copyright (c), 2016-2019, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -8,16 +7,9 @@
 #
 # @author Davide Brunato <brunato@sissa.it>
 #
-"""
-This module contains exception and warning classes for the 'xmlschema.validators' subpackage.
-"""
-from __future__ import unicode_literals
-
-from ..compat import PY3
 from ..exceptions import XMLSchemaException, XMLSchemaWarning, XMLSchemaValueError
-from ..etree import etree_tostring, is_etree_element, etree_getpath
-from ..helpers import qname_to_prefixed
-from ..resources import XMLResource
+from ..qnames import get_prefixed_qname
+from ..etree import is_etree_element, etree_tostring, etree_getpath
 
 
 class XMLSchemaValidatorError(XMLSchemaException):
@@ -34,58 +26,50 @@ class XMLSchemaValidatorError(XMLSchemaException):
     :type source: XMLResource
     :param namespaces: is an optional mapping from namespace prefix to URI.
     :type namespaces: dict
-    :ivar path: the XPath of the element, calculated when the element is set or the XML resource is set.
+    :ivar path: the XPath of the element, calculated when the element is set \
+    or the XML resource is set.
     """
     def __init__(self, validator, message, elem=None, source=None, namespaces=None):
+        self.path = None
         self.validator = validator
         self.message = message[:-1] if message[-1] in ('.', ':') else message
         self.namespaces = namespaces
-        self.elem = elem
         self.source = source
+        self.elem = elem
 
     def __str__(self):
-        return unicode(self).encode("utf-8")
-
-    def __unicode__(self):
         if self.elem is None:
-            return '%s.' % self.message
-        else:
-            msg = ['%s:\n' % self.message]
-            if self.elem is not None:
-                elem_as_string = etree_tostring(self.elem, self.namespaces, '  ', 20)
-                if hasattr(self.elem, 'sourceline'):
-                    msg.append("Schema (line %r):\n\n%s\n" % (self.elem.sourceline, elem_as_string))
-                else:
-                    msg.append("Schema:\n\n%s\n" % elem_as_string)
-            if self.path is not None:
-                msg.append("Path: %s\n" % self.path)
-            if self.schema_url is not None:
-                msg.append("Schema URL: %s\n" % self.schema_url)
-                if self.origin_url not in (None, self.schema_url):
-                    msg.append("Origin URL: %s\n" % self.origin_url)
-            return '\n'.join(msg)
+            return self.message
 
-    if PY3:
-        __str__ = __unicode__
+        msg = ['%s:\n' % self.message]
+        elem_as_string = etree_tostring(self.elem, self.namespaces, '  ', 20)
+        msg.append("Schema:\n\n%s\n" % elem_as_string)
+
+        if self.path is not None:
+            msg.append("Path: %s\n" % self.path)
+        if self.schema_url is not None:
+            msg.append("Schema URL: %s\n" % self.schema_url)
+            if self.origin_url not in (None, self.schema_url):
+                msg.append("Origin URL: %s\n" % self.origin_url)
+        return '\n'.join(msg)
 
     @property
     def msg(self):
-        return self.__unicode__()
+        return self.__str__()
 
     def __setattr__(self, name, value):
-        if name == 'elem' and value is not None and not is_etree_element(value):
-            raise XMLSchemaValueError("'elem' attribute requires an Element, not %r." % type(value))
+        if name == 'elem' and value is not None:
+            if not is_etree_element(value):
+                raise XMLSchemaValueError(
+                    "'elem' attribute requires an Element, not %r." % type(value)
+                )
+            if self.source is not None:
+                self.path = etree_getpath(value, self.root, self.namespaces,
+                                          relative=False, add_position=True)
+                if self.source.is_lazy():
+                    value = None  # Don't save the element of a lazy resource
         super(XMLSchemaValidatorError, self).__setattr__(name, value)
 
-        # Calculate and set the element's path: have to be calculated asap because is the
-        # XML resource is lazy the intermediate nodes could be deleted.
-        if name in ('elem', 'source'):
-            elem, root = self.elem, self.root
-            if not is_etree_element(elem) or not is_etree_element(root):
-                self.path = None
-            else:
-                self.path = etree_getpath(elem, root, self.namespaces, relative=False, add_position=True)
-
     @property
     def sourceline(self):
         return getattr(self.elem, 'sourceline', None)
@@ -174,7 +158,7 @@ class XMLSchemaModelError(XMLSchemaValidatorError, ValueError):
 class XMLSchemaModelDepthError(XMLSchemaModelError):
     """Raised when recursion depth is exceeded while iterating a model group."""
     def __init__(self, group):
-        msg = "maximum model recursion depth exceeded while iterating group %r" % group
+        msg = "maximum model recursion depth exceeded while iterating {!r}".format(group)
         super(XMLSchemaModelDepthError, self).__init__(group, message=msg)
 
 
@@ -197,9 +181,14 @@ class XMLSchemaValidationError(XMLSchemaValidatorError, ValueError):
     :type namespaces: dict
     """
     def __init__(self, validator, obj, reason=None, source=None, namespaces=None):
+        if not isinstance(obj, str):
+            _obj = obj
+        else:
+            _obj = obj.encode('ascii', 'xmlcharrefreplace').decode('utf-8')
+
         super(XMLSchemaValidationError, self).__init__(
             validator=validator,
-            message="failed validating {!r} with {!r}".format(obj, validator),
+            message="failed validating {!r} with {!r}".format(_obj, validator),
             elem=obj if is_etree_element(obj) else None,
             source=source,
             namespaces=namespaces,
@@ -207,29 +196,31 @@ class XMLSchemaValidationError(XMLSchemaValidatorError, ValueError):
         self.obj = obj
         self.reason = reason
 
-    def __str__(self):
-        # noinspection PyCompatibility,PyUnresolvedReferences
-        return unicode(self).encode("utf-8")
+    def __repr__(self):
+        return '%s(reason=%r)' % (self.__class__.__name__, self.reason)
 
-    def __unicode__(self):
+    def __str__(self):
         msg = ['%s:\n' % self.message]
         if self.reason is not None:
             msg.append('Reason: %s\n' % self.reason)
         if hasattr(self.validator, 'tostring'):
             msg.append("Schema:\n\n%s\n" % self.validator.tostring('  ', 20))
-        if self.elem is not None:
-            elem_as_string = etree_tostring(self.elem, self.namespaces, '  ', 20)
+        if is_etree_element(self.elem):
+            try:
+                elem_as_string = etree_tostring(self.elem, self.namespaces, '  ', 20)
+            except (ValueError, TypeError):        # pragma: no cover
+                elem_as_string = repr(self.elem)   # pragma: no cover
+
             if hasattr(self.elem, 'sourceline'):
                 msg.append("Instance (line %r):\n\n%s\n" % (self.elem.sourceline, elem_as_string))
             else:
                 msg.append("Instance:\n\n%s\n" % elem_as_string)
         if self.path is not None:
             msg.append("Path: %s\n" % self.path)
+        if len(msg) == 1:
+            return msg[0][:-2]
         return '\n'.join(msg)
 
-    if PY3:
-        __str__ = __unicode__
-
 
 class XMLSchemaDecodeError(XMLSchemaValidationError):
     """
@@ -300,17 +291,18 @@ class XMLSchemaChildrenValidationError(XMLSchemaValidationError):
     :param namespaces: is an optional mapping from namespace prefix to URI.
     :type namespaces: dict
     """
-    def __init__(self, validator, elem, index, particle, occurs=0, expected=None, source=None, namespaces=None):
+    def __init__(self, validator, elem, index, particle, occurs=0,
+                 expected=None, source=None, namespaces=None):
         self.index = index
         self.particle = particle
         self.occurs = occurs
         self.expected = expected
 
-        tag = qname_to_prefixed(elem.tag, validator.namespaces)
+        tag = get_prefixed_qname(elem.tag, validator.namespaces, use_empty=False)
         if index >= len(elem):
             reason = "The content of element %r is not complete." % tag
         else:
-            child_tag = qname_to_prefixed(elem[index].tag, validator.namespaces)
+            child_tag = get_prefixed_qname(elem[index].tag, validator.namespaces, use_empty=False)
             reason = "Unexpected child with tag %r at position %d." % (child_tag, index + 1)
 
         if occurs and particle.is_missing(occurs):
@@ -328,18 +320,19 @@ class XMLSchemaChildrenValidationError(XMLSchemaValidationError):
             expected_tags = []
             for xsd_element in expected:
                 if xsd_element.name is not None:
-                    expected_tags.append(repr(xsd_element.prefixed_name))
+                    expected_tags.append(xsd_element.prefixed_name)
                 elif xsd_element.process_contents == 'strict':
                     expected_tags.append('from %r namespace/s' % xsd_element.namespace)
 
             if not expected_tags:
-                reason += " No child element is expected at this point."
-            elif len(expected_tags) > 1:
-                reason += " Tags %s are expected." % expected_tags
-            else:
+                pass
+            elif len(expected_tags) == 1:
                 reason += " Tag %s expected." % expected_tags[0]
+            else:
+                reason += " Tag (%s) expected." % ' | '.join(expected_tags)
 
-        super(XMLSchemaChildrenValidationError, self).__init__(validator, elem, reason, source, namespaces)
+        super(XMLSchemaChildrenValidationError, self).\
+            __init__(validator, elem, reason, source, namespaces)
 
 
 class XMLSchemaIncludeWarning(XMLSchemaWarning):
@@ -348,3 +341,7 @@ class XMLSchemaIncludeWarning(XMLSchemaWarning):
 
 class XMLSchemaImportWarning(XMLSchemaWarning):
     """A schema namespace import fails."""
+
+
+class XMLSchemaTypeTableWarning(XMLSchemaWarning):
+    """Not equivalent type table found in model."""
